// This file has been autogenerated from a class added in the UI designer.

using System;

using Foundation;
using UIKit;
using ObjCRuntime;
using CoreAnimation;
using Metal;
using CoreGraphics;

namespace VixarEngine
{
	public partial class ImageView : UIView
	{
		[Export ("layerClass")]
		public static Class LayerClass ()
		{
			return new Class (typeof(CAMetalLayer));
		}

		IMTLDevice _device;
		MTLRenderPassDescriptor _renderPassDescriptor;
		CAMetalLayer _metalLayer;
		bool _layerSizeDidUpdate;

		IMTLTexture _depthTex;
		IMTLTexture _stencilTex;
		IMTLTexture _msaaTex;

		public MTLPixelFormat DepthPixelFormat { get; set; }

		public MTLPixelFormat StencilPixelFormat { get; set; }

		public nuint SampleCount { get; set; }

		public Renderer Renderer { get; set; }

		public override nfloat ContentScaleFactor {
			get {
				return base.ContentScaleFactor;
			}
			set {
				base.ContentScaleFactor = value;
				_layerSizeDidUpdate = true;
			}
		}

		public ImageView (IntPtr handle) : base (handle)
		{
			Opaque = true;
			BackgroundColor = UIColor.Clear;

			_metalLayer = (CAMetalLayer)Layer;

			_device = MTLDevice.SystemDefault;

			_metalLayer.Device = _device;
			_metalLayer.PixelFormat = MTLPixelFormat.BGRA8Unorm;
			_metalLayer.FramebufferOnly = true;
		}

		public ICAMetalDrawable GetNextDrawable ()
		{
			ICAMetalDrawable currentDrawable = null;

			currentDrawable = _metalLayer.NextDrawable ();
			if (currentDrawable == null)
				Console.WriteLine ("CurrentDrawable is null");

			return currentDrawable;
		}

		public void Display ()
		{
			if (_layerSizeDidUpdate) {
				CGSize drawableSize = Bounds.Size;
				drawableSize.Width *= ContentScaleFactor;
				drawableSize.Height *= ContentScaleFactor;
				_metalLayer.DrawableSize = drawableSize;
				_layerSizeDidUpdate = false;
			}

			Renderer.Render (this);
		}

		public MTLRenderPassDescriptor GetRenderPassDescriptor (ICAMetalDrawable drawable)
		{
			if (drawable == null) {
				Console.WriteLine ("ERROR: Failed to get a drawable!");
				_renderPassDescriptor = null;
			} else {
				SetupRenderPassDescriptorForTexture (drawable.Texture);
			}
			return _renderPassDescriptor;
		}

		public override void MovedToWindow ()
		{
			ContentScaleFactor = Window.Screen.NativeScale;
		}

		public override void LayoutSubviews ()
		{
			base.LayoutSubviews ();
			_layerSizeDidUpdate = true;
		}

		void SetupRenderPassDescriptorForTexture (IMTLTexture texture)
		{
			if (_renderPassDescriptor == null)
				_renderPassDescriptor = new MTLRenderPassDescriptor ();

			MTLRenderPassColorAttachmentDescriptor colorAttachment = _renderPassDescriptor.ColorAttachments [0];
			colorAttachment.Texture = texture;
			colorAttachment.LoadAction = MTLLoadAction.Clear;
			colorAttachment.ClearColor = new MTLClearColor (0.65f, 0.65f, 0.65f, 1.0f);

			if (SampleCount > 1) {
				if (_msaaTex == null || NeedUpdate (texture, _msaaTex)) {
					var desc = MTLTextureDescriptor.CreateTexture2DDescriptor (MTLPixelFormat.BGRA8Unorm, texture.Width, texture.Height, false);
					desc.TextureType = MTLTextureType.k2DMultisample;
					desc.SampleCount = SampleCount;
					_msaaTex = _device.CreateTexture (desc);
				}

				colorAttachment.Texture = _msaaTex;
				colorAttachment.ResolveTexture = texture;

				// set store action to resolve in this case
				colorAttachment.StoreAction = MTLStoreAction.MultisampleResolve;
			} else {
				colorAttachment.StoreAction = MTLStoreAction.Store;
			}

			if (DepthPixelFormat != MTLPixelFormat.Invalid && (_depthTex == null || NeedUpdate (texture, _depthTex))) {
					var desc = MTLTextureDescriptor.CreateTexture2DDescriptor (DepthPixelFormat, texture.Width, texture.Height, false);
					desc.TextureType = (SampleCount > 1) ? MTLTextureType.k2DMultisample : MTLTextureType.k2D;
					desc.SampleCount = SampleCount;
					_depthTex = _device.CreateTexture (desc);

					MTLRenderPassDepthAttachmentDescriptor depthAttachment = _renderPassDescriptor.DepthAttachment;
					depthAttachment.Texture = _depthTex;
					depthAttachment.LoadAction = MTLLoadAction.Clear;
					depthAttachment.StoreAction = MTLStoreAction.DontCare;
					depthAttachment.ClearDepth = 1.0;
			}

			if (StencilPixelFormat != MTLPixelFormat.Invalid && (_stencilTex == null || NeedUpdate (texture, _stencilTex))) {
					var desc = MTLTextureDescriptor.CreateTexture2DDescriptor (StencilPixelFormat, texture.Width, texture.Height, false);
					desc.TextureType = (SampleCount > 1) ? MTLTextureType.k2DMultisample : MTLTextureType.k2D;
					desc.SampleCount = SampleCount;
					_stencilTex = _device.CreateTexture (desc);

					MTLRenderPassStencilAttachmentDescriptor stencilAttachment = _renderPassDescriptor.StencilAttachment;
					stencilAttachment.Texture = _stencilTex;
					stencilAttachment.LoadAction = MTLLoadAction.Clear;
					stencilAttachment.StoreAction = MTLStoreAction.DontCare;
					stencilAttachment.ClearStencil = 0;
			}
		}

		bool NeedUpdate (IMTLTexture texture, IMTLTexture textureToCompare)
		{
			bool doUpdate = false;
			doUpdate = (textureToCompare.Width != texture.Width) || (textureToCompare.Height != texture.Height) || (textureToCompare.SampleCount != SampleCount);

			return doUpdate;
		}
	}
}
